@@ -1,6 +1,21 @@
+/*
+ * Copyright 2014 - Present Rafael Winterhalter
+ *
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
+ *
+ *     http://www.apache.org/licenses/LICENSE-2.0
+ *
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
+ */
 package net.bytebuddy.implementation;
 
-import net.bytebuddy.ClassFileVersion;
+import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;
 import net.bytebuddy.build.HashCodeAndEqualsPlugin;
 import net.bytebuddy.description.field.FieldDescription;
 import net.bytebuddy.description.method.MethodDescription;
@@ -21,10 +36,8 @@ import net.bytebuddy.matcher.ElementMatcher;
 import org.objectweb.asm.Label;
 import org.objectweb.asm.MethodVisitor;
 import org.objectweb.asm.Opcodes;
-import org.objectweb.asm.Type;
 
-import java.util.ArrayList;
-import java.util.List;
+import java.util.*;
 
 import static net.bytebuddy.matcher.ElementMatchers.*;
 
@@ -40,7 +53,7 @@ public class EqualsMethod implements Implementation {
     /**
      * The {@link Object#equals(Object)} method.
      */
-    private static final MethodDescription.InDefinedShape EQUALS = TypeDescription.OBJECT
+    private static final MethodDescription.InDefinedShape EQUALS = TypeDescription.ForLoadedType.of(Object.class)
             .getDeclaredMethods()
             .filter(isEquals())
             .getOnly();
@@ -65,13 +78,23 @@ public class EqualsMethod implements Implementation {
      */
     private final ElementMatcher.Junction<? super FieldDescription.InDefinedShape> nonNullable;
 
+    /**
+     * A matcher to determine that a field should be considered by its identity.
+     */
+    private final ElementMatcher.Junction<? super FieldDescription.InDefinedShape> identity;
+
+    /**
+     * The comparator to apply for ordering fields.
+     */
+    private final Comparator<? super FieldDescription.InDefinedShape> comparator;
+
     /**
      * Creates a new equals method implementation.
      *
      * @param superClassCheck The baseline equality to check.
      */
     protected EqualsMethod(SuperClassCheck superClassCheck) {
-        this(superClassCheck, TypeCompatibilityCheck.EXACT, none(), none());
+        this(superClassCheck, TypeCompatibilityCheck.EXACT, none(), none(), none(), NaturalOrderComparator.INSTANCE);
     }
 
     /**
@@ -81,15 +104,21 @@ public class EqualsMethod implements Implementation {
      * @param typeCompatibilityCheck The instance type compatibility check.
      * @param ignored                A matcher to filter fields that should not be used for a equality resolution.
      * @param nonNullable            A matcher to determine fields of a reference type that cannot be {@code null}.
+     * @param identity               A matcher to determine that a field should be considered by its identity.
+     * @param comparator             The comparator to apply for ordering fields.
      */
     private EqualsMethod(SuperClassCheck superClassCheck,
                          TypeCompatibilityCheck typeCompatibilityCheck,
                          ElementMatcher.Junction<? super FieldDescription.InDefinedShape> ignored,
-                         ElementMatcher.Junction<? super FieldDescription.InDefinedShape> nonNullable) {
+                         ElementMatcher.Junction<? super FieldDescription.InDefinedShape> nonNullable,
+                         ElementMatcher.Junction<? super FieldDescription.InDefinedShape> identity,
+                         Comparator<? super FieldDescription.InDefinedShape> comparator) {
         this.superClassCheck = superClassCheck;
         this.typeCompatibilityCheck = typeCompatibilityCheck;
         this.ignored = ignored;
         this.nonNullable = nonNullable;
+        this.identity = identity;
+        this.comparator = comparator;
     }
 
     /**
@@ -118,7 +147,7 @@ public class EqualsMethod implements Implementation {
      * @return A new version of this equals method implementation that also ignores any fields matched by the provided matcher.
      */
     public EqualsMethod withIgnoredFields(ElementMatcher<? super FieldDescription.InDefinedShape> ignored) {
-        return new EqualsMethod(superClassCheck, typeCompatibilityCheck, this.ignored.<FieldDescription.InDefinedShape>or(ignored), nonNullable);
+        return new EqualsMethod(superClassCheck, typeCompatibilityCheck, this.ignored.<FieldDescription.InDefinedShape>or(ignored), nonNullable, identity, comparator);
     }
 
     /**
@@ -130,7 +159,65 @@ public class EqualsMethod implements Implementation {
      * the provided matcher.
      */
     public EqualsMethod withNonNullableFields(ElementMatcher<? super FieldDescription.InDefinedShape> nonNullable) {
-        return new EqualsMethod(superClassCheck, typeCompatibilityCheck, ignored, this.nonNullable.<FieldDescription.InDefinedShape>or(nonNullable));
+        return new EqualsMethod(superClassCheck, typeCompatibilityCheck, ignored, this.nonNullable.<FieldDescription.InDefinedShape>or(nonNullable), identity, comparator);
+    }
+
+    /**
+     * Returns a new version of this equals method implementation that considers the matched fields by their identity.
+     *
+     * @param identity A matcher to determine that a field should be considered by its identity.
+     * @return A new version of this equals method implementation that also considers the matched fields by their identity.
+     */
+    public EqualsMethod withIdentityFields(ElementMatcher<? super FieldDescription.InDefinedShape> identity) {
+        return new EqualsMethod(superClassCheck, typeCompatibilityCheck, ignored, nonNullable, this.identity.<FieldDescription.InDefinedShape>or(identity), comparator);
+    }
+
+    /**
+     * Returns a new version of this equals method that compares fields with primitive types prior to fields with non-primitive types.
+     *
+     * @return A new version of this equals method that compares primitive-typed fields before fields with non-primitive-typed fields.
+     */
+    public EqualsMethod withPrimitiveTypedFieldsFirst() {
+        return withFieldOrder(TypePropertyComparator.FOR_PRIMITIVE_TYPES);
+    }
+
+    /**
+     * Returns a new version of this equals method that compares fields with enumeration types prior to fields with non-enumeration types.
+     *
+     * @return A new version of this equals method that compares enumeration-typed fields before fields with non-enumeration-typed fields.
+     */
+    public EqualsMethod withEnumerationTypedFieldsFirst() {
+        return withFieldOrder(TypePropertyComparator.FOR_ENUMERATION_TYPES);
+    }
+
+    /**
+     * Returns a new version of this equals method that compares fields with primitive wrapper types prior to fields with non-primitive wrapper types.
+     *
+     * @return A new version of this equals method that compares primitive wrapper-typed fields before fields with non-primitive wrapper-typed fields.
+     */
+    public EqualsMethod withPrimitiveWrapperTypedFieldsFirst() {
+        return withFieldOrder(TypePropertyComparator.FOR_PRIMITIVE_WRAPPER_TYPES);
+    }
+
+    /**
+     * Returns a new version of this equals method that compares fields with {@link String} types prior to fields with non-{@link String} types.
+     *
+     * @return A new version of this equals method that compares {@link String}-typed fields before fields with non-{@link String}-typed fields.
+     */
+    public EqualsMethod withStringTypedFieldsFirst() {
+        return withFieldOrder(TypePropertyComparator.FOR_STRING_TYPES);
+    }
+
+    /**
+     * Applies the supplied comparator to determine an order for fields for being compared. Fields with the lowest sort order are compared
+     * first. Any previously defined comparators are applied prior to the supplied comparator.
+     *
+     * @param comparator The comparator to apply.
+     * @return A new version of this equals method that sorts fields in their application order using the supplied comparator.
+     */
+    @SuppressWarnings("unchecked") // In absence of @SafeVarargs
+    public EqualsMethod withFieldOrder(Comparator<? super FieldDescription.InDefinedShape> comparator) {
+        return new EqualsMethod(superClassCheck, typeCompatibilityCheck, ignored, nonNullable, identity, new CompoundComparator(this.comparator, comparator));
     }
 
     /**
@@ -141,26 +228,34 @@ public class EqualsMethod implements Implementation {
      * of the instrumented type instead of requiring an exact match.
      */
     public Implementation withSubclassEquality() {
-        return new EqualsMethod(superClassCheck, TypeCompatibilityCheck.SUBCLASS, ignored, nonNullable);
+        return new EqualsMethod(superClassCheck, TypeCompatibilityCheck.SUBCLASS, ignored, nonNullable, identity, comparator);
     }
 
-    @Override
+    /**
+     * {@inheritDoc}
+     */
     public InstrumentedType prepare(InstrumentedType instrumentedType) {
         return instrumentedType;
     }
 
-    @Override
+    /**
+     * {@inheritDoc}
+     */
     public ByteCodeAppender appender(Target implementationTarget) {
         if (implementationTarget.getInstrumentedType().isInterface()) {
             throw new IllegalStateException("Cannot implement meaningful equals method for " + implementationTarget.getInstrumentedType());
         }
+        List<FieldDescription.InDefinedShape> fields = new ArrayList<FieldDescription.InDefinedShape>(implementationTarget.getInstrumentedType()
+                .getDeclaredFields()
+                .filter(not(isStatic().or(ignored))));
+        Collections.sort(fields, comparator);
         return new Appender(implementationTarget.getInstrumentedType(), new StackManipulation.Compound(
                 superClassCheck.resolve(implementationTarget.getInstrumentedType()),
                 MethodVariableAccess.loadThis(),
                 MethodVariableAccess.REFERENCE.loadFrom(1),
                 ConditionalReturn.onIdentity().returningTrue(),
                 typeCompatibilityCheck.resolve(implementationTarget.getInstrumentedType())
-        ), implementationTarget.getInstrumentedType().getDeclaredFields().filter(not(isStatic().or(ignored))), nonNullable);
+        ), fields, nonNullable, identity);
     }
 
     /**
@@ -294,17 +389,23 @@ public class EqualsMethod implements Implementation {
              */
             INSTANCE;
 
-            @Override
+            /**
+             * {@inheritDoc}
+             */
             public StackManipulation before() {
                 return StackManipulation.Trivial.INSTANCE;
             }
 
-            @Override
+            /**
+             * {@inheritDoc}
+             */
             public StackManipulation after() {
                 return StackManipulation.Trivial.INSTANCE;
             }
 
-            @Override
+            /**
+             * {@inheritDoc}
+             */
             public int getRequiredVariablePadding() {
                 return StackSize.ZERO.getSize();
             }
@@ -317,16 +418,6 @@ public class EqualsMethod implements Implementation {
         @HashCodeAndEqualsPlugin.Enhance
         class UsingJump implements NullValueGuard {
 
-            /**
-             * An empty array.
-             */
-            private static final Object[] EMPTY = new Object[0];
-
-            /**
-             * An array containing a single reference value.
-             */
-            private static final Object[] REFERENCE = new Object[]{Type.getInternalName(Object.class)};
-
             /**
              * The instrumented method.
              */
@@ -359,17 +450,23 @@ public class EqualsMethod implements Implementation {
                 endOfBlock = new Label();
             }
 
-            @Override
+            /**
+             * {@inheritDoc}
+             */
             public StackManipulation before() {
                 return new UsingJump.BeforeInstruction();
             }
 
-            @Override
+            /**
+             * {@inheritDoc}
+             */
             public StackManipulation after() {
                 return new UsingJump.AfterInstruction();
             }
 
-            @Override
+            /**
+             * {@inheritDoc}
+             */
             public int getRequiredVariablePadding() {
                 return 2;
             }
@@ -378,14 +475,11 @@ public class EqualsMethod implements Implementation {
              * The stack manipulation to apply before the equality computation.
              */
             @HashCodeAndEqualsPlugin.Enhance(includeSyntheticFields = true)
-            protected class BeforeInstruction implements StackManipulation {
+            protected class BeforeInstruction extends StackManipulation.AbstractBase {
 
-                @Override
-                public boolean isValid() {
-                    return true;
-                }
-
-                @Override
+                /**
+                 * {@inheritDoc}
+                 */
                 public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                     methodVisitor.visitVarInsn(Opcodes.ASTORE, instrumentedMethod.getStackSize());
                     methodVisitor.visitVarInsn(Opcodes.ASTORE, instrumentedMethod.getStackSize() + 1);
@@ -395,7 +489,7 @@ public class EqualsMethod implements Implementation {
                     methodVisitor.visitJumpInsn(Opcodes.IFNULL, firstValueNull);
                     methodVisitor.visitVarInsn(Opcodes.ALOAD, instrumentedMethod.getStackSize() + 1);
                     methodVisitor.visitVarInsn(Opcodes.ALOAD, instrumentedMethod.getStackSize());
-                    return new Size(0, 0);
+                    return Size.ZERO;
                 }
             }
 
@@ -403,32 +497,27 @@ public class EqualsMethod implements Implementation {
              * The stack manipulation to apply after the equality computation.
              */
             @HashCodeAndEqualsPlugin.Enhance(includeSyntheticFields = true)
-            protected class AfterInstruction implements StackManipulation {
+            protected class AfterInstruction extends StackManipulation.AbstractBase {
 
-                @Override
-                public boolean isValid() {
-                    return true;
-                }
-
-                @Override
+                /**
+                 * {@inheritDoc}
+                 */
                 public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                     methodVisitor.visitJumpInsn(Opcodes.GOTO, endOfBlock);
                     methodVisitor.visitLabel(secondValueNull);
-                    if (implementationContext.getClassFileVersion().isAtLeast(ClassFileVersion.JAVA_V6)) {
-                        methodVisitor.visitFrame(Opcodes.F_SAME1, EMPTY.length, EMPTY, REFERENCE.length, REFERENCE);
-                    }
+                    implementationContext.getFrameGeneration().same1(methodVisitor,
+                            TypeDescription.ForLoadedType.of(Object.class),
+                            Arrays.asList(implementationContext.getInstrumentedType(), TypeDescription.ForLoadedType.of(Object.class)));
                     methodVisitor.visitJumpInsn(Opcodes.IFNULL, endOfBlock);
                     methodVisitor.visitLabel(firstValueNull);
-                    if (implementationContext.getClassFileVersion().isAtLeast(ClassFileVersion.JAVA_V6)) {
-                        methodVisitor.visitFrame(Opcodes.F_SAME, EMPTY.length, EMPTY, EMPTY.length, EMPTY);
-                    }
+                    implementationContext.getFrameGeneration().same(methodVisitor,
+                            Arrays.asList(implementationContext.getInstrumentedType(), TypeDescription.ForLoadedType.of(Object.class)));
                     methodVisitor.visitInsn(Opcodes.ICONST_0);
                     methodVisitor.visitInsn(Opcodes.IRETURN);
                     methodVisitor.visitLabel(endOfBlock);
-                    if (implementationContext.getClassFileVersion().isAtLeast(ClassFileVersion.JAVA_V6)) {
-                        methodVisitor.visitFrame(Opcodes.F_SAME, EMPTY.length, EMPTY, EMPTY.length, EMPTY);
-                    }
-                    return new Size(0, 0);
+                    implementationContext.getFrameGeneration().same(methodVisitor,
+                            Arrays.asList(implementationContext.getInstrumentedType(), TypeDescription.ForLoadedType.of(Object.class)));
+                    return Size.ZERO;
                 }
             }
         }
@@ -443,7 +532,7 @@ public class EqualsMethod implements Implementation {
          * A comparator for a {@code long} value.
          */
         LONG {
-            @Override
+            /** {@inheritDoc} */
             public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                 methodVisitor.visitInsn(Opcodes.LCMP);
                 return new Size(-2, 0);
@@ -454,9 +543,9 @@ public class EqualsMethod implements Implementation {
          * A comparator for a {@code float} value.
          */
         FLOAT {
-            @Override
+            /** {@inheritDoc} */
             public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
-                methodVisitor.visitInsn(Opcodes.FCMPL);
+                methodVisitor.visitMethodInsn(Opcodes.INVOKESTATIC, "java/lang/Float", "compare", "(FF)I", false);
                 return new Size(-1, 0);
             }
         },
@@ -465,9 +554,9 @@ public class EqualsMethod implements Implementation {
          * A comparator for a {@code double} value.
          */
         DOUBLE {
-            @Override
+            /** {@inheritDoc} */
             public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
-                methodVisitor.visitInsn(Opcodes.DCMPL);
+                methodVisitor.visitMethodInsn(Opcodes.INVOKESTATIC, "java/lang/Double", "compare", "(DD)I", false);
                 return new Size(-2, 0);
             }
         },
@@ -476,7 +565,7 @@ public class EqualsMethod implements Implementation {
          * A comparator for a {@code boolean[]} value.
          */
         BOOLEAN_ARRAY {
-            @Override
+            /** {@inheritDoc} */
             public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                 methodVisitor.visitMethodInsn(Opcodes.INVOKESTATIC, "java/util/Arrays", "equals", "([Z[Z)Z", false);
                 return new Size(-1, 0);
@@ -487,7 +576,7 @@ public class EqualsMethod implements Implementation {
          * A comparator for a {@code byte[]} value.
          */
         BYTE_ARRAY {
-            @Override
+            /** {@inheritDoc} */
             public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                 methodVisitor.visitMethodInsn(Opcodes.INVOKESTATIC, "java/util/Arrays", "equals", "([B[B)Z", false);
                 return new Size(-1, 0);
@@ -498,7 +587,7 @@ public class EqualsMethod implements Implementation {
          * A comparator for a {@code short[]} value.
          */
         SHORT_ARRAY {
-            @Override
+            /** {@inheritDoc} */
             public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                 methodVisitor.visitMethodInsn(Opcodes.INVOKESTATIC, "java/util/Arrays", "equals", "([S[S)Z", false);
                 return new Size(-1, 0);
@@ -509,7 +598,7 @@ public class EqualsMethod implements Implementation {
          * A comparator for a {@code char[]} value.
          */
         CHARACTER_ARRAY {
-            @Override
+            /** {@inheritDoc} */
             public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                 methodVisitor.visitMethodInsn(Opcodes.INVOKESTATIC, "java/util/Arrays", "equals", "([C[C)Z", false);
                 return new Size(-1, 0);
@@ -520,7 +609,7 @@ public class EqualsMethod implements Implementation {
          * A comparator for an {@code int[]} value.
          */
         INTEGER_ARRAY {
-            @Override
+            /** {@inheritDoc} */
             public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                 methodVisitor.visitMethodInsn(Opcodes.INVOKESTATIC, "java/util/Arrays", "equals", "([I[I)Z", false);
                 return new Size(-1, 0);
@@ -531,7 +620,7 @@ public class EqualsMethod implements Implementation {
          * A comparator for a {@code long[]} value.
          */
         LONG_ARRAY {
-            @Override
+            /** {@inheritDoc} */
             public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                 methodVisitor.visitMethodInsn(Opcodes.INVOKESTATIC, "java/util/Arrays", "equals", "([J[J)Z", false);
                 return new Size(-1, 0);
@@ -542,7 +631,7 @@ public class EqualsMethod implements Implementation {
          * A comparator for a {@code float[]} value.
          */
         FLOAT_ARRAY {
-            @Override
+            /** {@inheritDoc} */
             public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                 methodVisitor.visitMethodInsn(Opcodes.INVOKESTATIC, "java/util/Arrays", "equals", "([F[F)Z", false);
                 return new Size(-1, 0);
@@ -553,7 +642,7 @@ public class EqualsMethod implements Implementation {
          * A transformer for a {@code double[]} value.
          */
         DOUBLE_ARRAY {
-            @Override
+            /** {@inheritDoc} */
             public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                 methodVisitor.visitMethodInsn(Opcodes.INVOKESTATIC, "java/util/Arrays", "equals", "([D[D)Z", false);
                 return new Size(-1, 0);
@@ -564,7 +653,7 @@ public class EqualsMethod implements Implementation {
          * A transformer for a reference array value.
          */
         REFERENCE_ARRAY {
-            @Override
+            /** {@inheritDoc} */
             public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                 methodVisitor.visitMethodInsn(Opcodes.INVOKESTATIC, "java/util/Arrays", "equals", "([Ljava/lang/Object;[Ljava/lang/Object;)Z", false);
                 return new Size(-1, 0);
@@ -575,7 +664,7 @@ public class EqualsMethod implements Implementation {
          * A transformer for a nested reference array value.
          */
         NESTED_ARRAY {
-            @Override
+            /** {@inheritDoc} */
             public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
                 methodVisitor.visitMethodInsn(Opcodes.INVOKESTATIC, "java/util/Arrays", "deepEquals", "([Ljava/lang/Object;[Ljava/lang/Object;)Z", false);
                 return new Size(-1, 0);
@@ -588,6 +677,7 @@ public class EqualsMethod implements Implementation {
          * @param typeDefinition The type definition to resolve.
          * @return The stack manipulation to apply.
          */
+        @SuppressFBWarnings(value = "NP_NULL_ON_SOME_PATH_FROM_RETURN_VALUE", justification = "Assuming component type for array type.")
         public static StackManipulation of(TypeDefinition typeDefinition) {
             if (typeDefinition.represents(boolean.class)
                     || typeDefinition.represents(byte.class)
@@ -626,7 +716,9 @@ public class EqualsMethod implements Implementation {
             }
         }
 
-        @Override
+        /**
+         * {@inheritDoc}
+         */
         public boolean isValid() {
             return true;
         }
@@ -658,6 +750,11 @@ public class EqualsMethod implements Implementation {
          */
         private final ElementMatcher<? super FieldDescription.InDefinedShape> nonNullable;
 
+        /**
+         * A matcher to determine fields of a reference type that cannot be {@code null}.
+         */
+        private final ElementMatcher<? super FieldDescription.InDefinedShape> identity;
+
         /**
          * Creates a new appender.
          *
@@ -665,18 +762,23 @@ public class EqualsMethod implements Implementation {
          * @param baseline          The baseline stack manipulation.
          * @param fieldDescriptions A list of fields to use for the comparison.
          * @param nonNullable       A matcher to determine fields of a reference type that cannot be {@code null}.
+         * @param identity          A matcher to determine that a field should be considered by its identity.
          */
         protected Appender(TypeDescription instrumentedType,
                            StackManipulation baseline,
                            List<FieldDescription.InDefinedShape> fieldDescriptions,
-                           ElementMatcher<? super FieldDescription.InDefinedShape> nonNullable) {
+                           ElementMatcher<? super FieldDescription.InDefinedShape> nonNullable,
+                           ElementMatcher<? super FieldDescription.InDefinedShape> identity) {
             this.instrumentedType = instrumentedType;
             this.baseline = baseline;
             this.fieldDescriptions = fieldDescriptions;
             this.nonNullable = nonNullable;
+            this.identity = identity;
         }
 
-        @Override
+        /**
+         * {@inheritDoc}
+         */
         public Size apply(MethodVisitor methodVisitor, Context implementationContext, MethodDescription instrumentedMethod) {
             if (instrumentedMethod.isStatic()) {
                 throw new IllegalStateException("Hash code method must not be static: " + instrumentedMethod);
@@ -694,13 +796,17 @@ public class EqualsMethod implements Implementation {
                 stackManipulations.add(MethodVariableAccess.REFERENCE.loadFrom(1));
                 stackManipulations.add(TypeCasting.to(instrumentedType));
                 stackManipulations.add(FieldAccess.forField(fieldDescription).read());
-                NullValueGuard nullValueGuard = fieldDescription.getType().isPrimitive() || fieldDescription.getType().isArray() || nonNullable.matches(fieldDescription)
-                        ? NullValueGuard.NoOp.INSTANCE
-                        : new NullValueGuard.UsingJump(instrumentedMethod);
-                stackManipulations.add(nullValueGuard.before());
-                stackManipulations.add(ValueComparator.of(fieldDescription.getType()));
-                stackManipulations.add(nullValueGuard.after());
-                padding = Math.max(padding, nullValueGuard.getRequiredVariablePadding());
+                if (!fieldDescription.getType().isPrimitive() && identity.matches(fieldDescription)) {
+                    stackManipulations.add(ConditionalReturn.onNonIdentity());
+                } else {
+                    NullValueGuard nullValueGuard = fieldDescription.getType().isPrimitive() || fieldDescription.getType().isArray() || nonNullable.matches(fieldDescription)
+                            ? NullValueGuard.NoOp.INSTANCE
+                            : new NullValueGuard.UsingJump(instrumentedMethod);
+                    stackManipulations.add(nullValueGuard.before());
+                    stackManipulations.add(ValueComparator.of(fieldDescription.getType()));
+                    stackManipulations.add(nullValueGuard.after());
+                    padding = Math.max(padding, nullValueGuard.getRequiredVariablePadding());
+                }
             }
             stackManipulations.add(IntegerConstant.forValue(true));
             stackManipulations.add(MethodReturn.INTEGER);
@@ -712,12 +818,7 @@ public class EqualsMethod implements Implementation {
      * A conditional return aborts the equality computation if a given condition was reached.
      */
     @HashCodeAndEqualsPlugin.Enhance
-    protected static class ConditionalReturn implements StackManipulation {
-
-        /**
-         * An empty array.
-         */
-        private static final Object[] EMPTY = new Object[0];
+    protected static class ConditionalReturn extends StackManipulation.AbstractBase {
 
         /**
          * The conditional jump instruction upon which the return is not triggered.
@@ -812,22 +913,155 @@ public class EqualsMethod implements Implementation {
             return new ConditionalReturn(jumpCondition, Opcodes.ICONST_1);
         }
 
-        @Override
-        public boolean isValid() {
-            return true;
-        }
-
-        @Override
+        /**
+         * {@inheritDoc}
+         */
         public Size apply(MethodVisitor methodVisitor, Context implementationContext) {
             Label label = new Label();
             methodVisitor.visitJumpInsn(jumpCondition, label);
             methodVisitor.visitInsn(value);
             methodVisitor.visitInsn(Opcodes.IRETURN);
             methodVisitor.visitLabel(label);
-            if (implementationContext.getClassFileVersion().isAtLeast(ClassFileVersion.JAVA_V6)) {
-                methodVisitor.visitFrame(Opcodes.F_SAME, EMPTY.length, EMPTY, EMPTY.length, EMPTY);
-            }
+            implementationContext.getFrameGeneration().same(methodVisitor,
+                    Arrays.asList(implementationContext.getInstrumentedType(), TypeDescription.ForLoadedType.of(Object.class)));
             return new Size(-1, 1);
         }
     }
+
+    /**
+     * A comparator that retains the natural order.
+     */
+    protected enum NaturalOrderComparator implements Comparator<FieldDescription.InDefinedShape> {
+
+        /**
+         * The singleton instance.
+         */
+        INSTANCE;
+
+        /**
+         * {@inheritDoc}
+         */
+        public int compare(FieldDescription.InDefinedShape left, FieldDescription.InDefinedShape right) {
+            return 0;
+        }
+    }
+
+    /**
+     * A comparator that sorts fields by a type property.
+     */
+    protected enum TypePropertyComparator implements Comparator<FieldDescription.InDefinedShape> {
+
+        /**
+         * Weights primitive types before non-primitive types.
+         */
+        FOR_PRIMITIVE_TYPES {
+            @Override
+            protected boolean resolve(TypeDefinition typeDefinition) {
+                return typeDefinition.isPrimitive();
+            }
+        },
+
+        /**
+         * Weights enumeration types before non-enumeration types.
+         */
+        FOR_ENUMERATION_TYPES {
+            @Override
+            protected boolean resolve(TypeDefinition typeDefinition) {
+                return typeDefinition.isEnum();
+            }
+        },
+
+        /**
+         * Weights {@link String} types first.
+         */
+        FOR_STRING_TYPES {
+            @Override
+            protected boolean resolve(TypeDefinition typeDefinition) {
+                return typeDefinition.represents(String.class);
+            }
+        },
+
+        /**
+         * Weights primitive wrapper types first.
+         */
+        FOR_PRIMITIVE_WRAPPER_TYPES {
+            @Override
+            protected boolean resolve(TypeDefinition typeDefinition) {
+                return typeDefinition.asErasure().isPrimitiveWrapper();
+            }
+        };
+
+        /**
+         * {@inheritDoc}
+         */
+        public int compare(FieldDescription.InDefinedShape left, FieldDescription.InDefinedShape right) {
+            if (resolve(left.getType()) && !resolve(right.getType())) {
+                return -1;
+            } else if (!resolve(left.getType()) && resolve(right.getType())) {
+                return 1;
+            } else {
+                return 0;
+            }
+        }
+
+        /**
+         * Resolves a type property.
+         *
+         * @param typeDefinition The type to resolve the property for.
+         * @return {@code true} if the type property is resolved.
+         */
+        protected abstract boolean resolve(TypeDefinition typeDefinition);
+    }
+
+    /**
+     * A compound comparator that compares the values of multiple fields.
+     */
+    @HashCodeAndEqualsPlugin.Enhance
+    @SuppressFBWarnings(value = "SE_COMPARATOR_SHOULD_BE_SERIALIZABLE", justification = "Not used within a serializable instance")
+    protected static class CompoundComparator implements Comparator<FieldDescription.InDefinedShape> {
+
+        /**
+         * All comparators to be applied in the application order.
+         */
+        private final List<Comparator<? super FieldDescription.InDefinedShape>> comparators;
+
+        /**
+         * Creates a compound comparator.
+         *
+         * @param comparator All comparators to be applied in the application order.
+         */
+        @SuppressWarnings("unchecked") // In absence of @SafeVarargs
+        protected CompoundComparator(Comparator<? super FieldDescription.InDefinedShape>... comparator) {
+            this(Arrays.asList(comparator));
+        }
+
+        /**
+         * Creates a compound comparator.
+         *
+         * @param comparators All comparators to be applied in the application order.
+         */
+        protected CompoundComparator(List<? extends Comparator<? super FieldDescription.InDefinedShape>> comparators) {
+            this.comparators = new ArrayList<Comparator<? super FieldDescription.InDefinedShape>>();
+            for (Comparator<? super FieldDescription.InDefinedShape> comparator : comparators) {
+                if (comparator instanceof CompoundComparator) {
+                    this.comparators.addAll(((CompoundComparator) comparator).comparators);
+                } else if (!(comparator instanceof NaturalOrderComparator)) {
+                    this.comparators.add(comparator);
+                }
+            }
+        }
+
+        /**
+         * {@inheritDoc}
+         */
+        public int compare(FieldDescription.InDefinedShape left, FieldDescription.InDefinedShape right) {
+            for (Comparator<? super FieldDescription.InDefinedShape> comparator : comparators) {
+                int comparison = comparator.compare(left, right);
+                if (comparison != 0) {
+                    return comparison;
+                }
+            }
+            return 0;
+        }
+    }
 }
